# -*- coding: utf-8 -*-
# Copyright: Alteo Gange <romanocaldoni@free.fr>
# License: GNU GPL, version 3 or later; http://www.gnu.org/copyleft/gpl.html
#
# This plugin allows you to easily memorize word sounds of Swac collections.
# There are currently more than 85000 words in fifteen languages.
# Look at informations() function below and the file swac/swac.html
#
# You must use a GNU/Linux distribution and to install the programs:
# python (>= 2.5), python-qt4 (>= 4.4.0), swac-get (>= 0.2), swac-play (>=0.1),
# anki (>= 0.9.9.6 <= 0.9.9.8 perhaps beyond)
# Read: http://shtooka.net/soft/swac-tools/en/

__version__ = "0.8.4-0" 
__date__ = _("August 2009")

from ankiqt import mw
from anki.hooks import runHook
from anki import stdmodels
from anki.utils import stripHTML, addTags, findTag, parseTags
from ankiqt.ui.modelchooser import AddModel, ModelChooser
from ankiqt import ui, appVersion
from ankiqt.ui.sound import getAudio
from anki.sound import play, clearAudioQueue
playOrigin = play

import os, sys, sqlite3, gettext, __builtin__, subprocess
from ConfigParser import SafeConfigParser
from functools import partial
from PyQt4.QtCore import *
from PyQt4.QtGui import *

# to generate the new file translation file ankiSwac.pot, execute the command:
# ~/.anki/plugins$ xgettext --no-wrap -o swac/locale/ankiSwac.pot swac.py
swacPluginResource = os.path.join(os.path.dirname(__file__), 'swac')
languageTrans = gettext.translation('ankiSwac',
                        os.path.join(swacPluginResource, "locale"),
                        languages=[mw.config["interfaceLang"]], fallback=True)
__builtin__.__dict__['_'] = partial(languageTrans.ugettext)

################### config parameters

swacDb = os.path.join(os.path.expanduser("~"), ".swac/swac.db")

# be carreful if you change the configDict key values because the Settings class
# depends of it (getattr() et setattr())!
configDict = {
    'swacModel': {              'value':_(u'Swac'),
                                'label':_("Swac model name: "),
                                'sort':0},
    'swacField': {              'value':_(u'Swac Expression'),
                                'label':_("Swac field name: "),
                                'sort':1},
    'phoneticField': {          'value':_(u'Phonetic'),
                                'label':_("Phonetic field name: "),
                                'sort':2},
    'homophonField': {          'value':_(u'Homophons'),
                                'label':_("Homophon field name: "),
                                'sort':3},
    'readingCardModel': {       'value':_(u'Reading'),
                                'label':_("Reading card name: "),
                                'sort':4},
    'recognitionCardModel': {   'value':_(u'Recognition'),
                                'label':_("Recognition card name: "),
                                'sort':5},
    'swacDatabasePath': {        'value':swacDb,
                                'label':_("Swac database file path: "),
                                'sort':6},
    'swacWordsSelectShortcut': {'value':_("F12"),
                                'label':_("Swac shortcut of expressions selection: "), 
                                'sort':7},
    'recordShortcut': {         'value':_("F10"),
                                'label':_("Swac shortcut to record your voice: "), 
                                'sort':8},
    'replayShortcut': {         'value':_("F11"),
                                'label':_("Swac shortcut to replay your voice: "), 
                                'sort':9},
    'viewPlayIcons': {          'value':_("0"),
                                'label':_("Hide the record and replay icons (0: no, other: yes): "), 
                                'sort':10},
}

class ConfigP(SafeConfigParser):
    "Basic configuration file parser language"
    loc = "UTF8" # locale
    defS = 'DEFAULTS' # defaults section
    def __init__(self):
        SafeConfigParser.__init__(self)
    def get(self, arg1, arg2=None):
        if arg2:
            return SafeConfigParser.get(self, arg1, arg2).decode(ConfigP.loc)
        else:
            return SafeConfigParser.get(self, ConfigP.defS, arg1).decode(ConfigP.loc)
    def set(self, arg1, arg2, arg3=None):
        if arg3:
            return SafeConfigParser.set(self, arg1, arg2, arg3.encode(ConfigP.loc))
        else:
            return SafeConfigParser.set(self, ConfigP.defS, arg1,
                                    arg2.encode(ConfigP.loc))
config = ConfigP()

configFilePath = os.path.join(swacPluginResource, "swac.conf")
config.read(configFilePath)
if not config.has_section(ConfigP.defS):
    config.add_section(ConfigP.defS)
for param in configDict.keys():
    if not config.has_option(ConfigP.defS, param) or not config.get(param):
        config.set(param, configDict[param]['value'])

# ISO 639-3 language symbol -> language
langDict = {'arb': _('Arabian'),
            'bel': _('Bielorussian'),
            'bul': _('Bulgarian'), # not yet
            'ces': _('Czech'),
            'chi': _('Chinese'),
            'deu': _('German'),
            'dut': _('Dutch'), # GPL
            'eng': _('English'),
            'fra': _('French'),
            'hun': _('Hungarian'), # not yet
            'ita': _('Italian'), # not yet
            'cmn': _('Mandarin'),
            'nld': _('Dutch'), # non GPL (and non free: "non libre" in French)
            'pol': _('Polish'), # not yet
            'por': _('Portuguese'),
            'rus': _('Russian'),
            'spa': _('Spanish'),
            'srp': _('Serbian'),
            'swe': _('Swedish'),
            'ukr': _('Ukrainian'),
            'wol': _('Wolof'),
        }

# flag to enable/disable the play button
ENABLE_PLAY_BUTTON = False

################### errors

def swacDataBaseError():
    QMessageBox.warning(mw, "Swac",
        _("Your Swac database is either not built or not valid"))

def swacBinaryError(binary):
    QMessageBox(mw, "Swac", _("You must install the following program: '%s' "
                "to enable the Swac functionalities") %", ".join(binary))

def severalSwacModelError():
    QMessageBox.warning(mw, "Swac",
        _("You mustn't have several models named \"%s\" in your deck")
                                                %config.get('swacModel'))

def langAnkiError():
    langs = ''
    for langSymbol, universalLang, localLang in languagesEnabled():
        if universalLang == localLang:
            langs += '\n%s : '%langSymbol + universalLang
        else:
            langs += '\n%s : '%langSymbol + universalLang +\
                        ' - ' + localLang
    QMessageBox.warning(mw, "Swac",
        _("You must add a valid symbol tag lang in your \"%(swacModel)s\" model.\n"
        "The valid symbols tags langs are currently:\n%(langs)s")
            %{'swacModel': config.get('swacModel'), 'langs': langs})

def langsAnkiError(symbolLangList):
    QMessageBox.warning(mw, "Swac",
        _("You mustn't have more than one symbol tag lang in your \"%(model)s\" "
        "model. You have currently the following symbols in your \"%(model)s\" "
        "model:\n%(symbolLang)s")
        %{'model': config.get('swacModel'), 'symbolLang': ", ".join(symbolLangList)})

def swacFieldError():
    QMessageBox.warning(mw, "Swac",
        _("You must have a \"%(field)s\" field in your \"%(model)s\" model")
            %{'field': config.get('swacField'), 'model': config.get('swacModel')})

def severalSwacFieldError():
     QMessageBox.warning(mw, "Swac",
        _("You mustn't have more than one field \"%(field)s\" in your \"%(model)s\" "
        "model")
            %{'field': config.get('swacField'), 'model': config.get('swacModel')})

def findAudioError(value):
    QMessageBox.warning(mw, "Swac",
        _("\"%(value)s\" value of the field \"%(field)s\" of the model "
            "\"%(model)s\" is not correct")
                %{'value': value, 'field': config.get('swacField'), 'model': config.get('swacModel')})

################### a few functions

def languagesEnabled():
    """Return the list:
    [(iso-639-3_symbol_lang, original name language, local name language), ...]
    ... related to the collections installed on the system or False if the Swac
    database doesn't exist
    ex: [(fra, Français, French), (swe, Svenska, Swedish)]"""
    languagesTuple = []
    langOriginalNameDict = {}
    langSymbolList = swacGetAll("""
select iso_639_3, title from languages
ORDER BY iso_639_3 ASC;""")
    # all languages
    for langSymbol, langOriginalName in langSymbolList:
        langOriginalNameDict[langSymbol] = langOriginalName
    currentLangSymbolList = swacGetOne("""
select packid from packages
ORDER BY packid ASC;""")
    uniqueLang = set()
    # languages related to the collections installed
    for packid in currentLangSymbolList:
        # ex: packid = fra-balm-voc
        langSymbol = packid.split('-')[0]
        if langSymbol in uniqueLang:
            continue
        else:
            uniqueLang.add(langSymbol)
        languagesTuple.append((langSymbol, langOriginalNameDict[langSymbol],
                                langDict.get(langSymbol,_('Unknown'))))
    if languagesTuple:
        return languagesTuple
    else:
        swacDataBaseError()
        return False

def splitSwacPattern(text):
    """Split the value 'expression (determinative) [intonation]'
    into expression, determinative and intonation"""
    expression, determinative, intonation = "", "", ""
    k = 0
    for start, end in (('(', ')'), ('[', ']')):
        try:
            i = text.index(start)
            j = text.index(end, i + 1)
            typ = text[i+1:j]
        except ValueError:
            typ = ""
        else:
            text = text[0:i] + text[j+1:len(text)]
        if k == 0:
            determinative = typ
        elif k== 1:
            intonation = typ
        k += 1
    expression = text
    return expression.strip(), determinative.strip(), intonation.strip()

def isBinary(bin):
    "Check if a binary is installed on the system"
    process = subprocess.Popen("which %s" %bin, stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT, shell=True)
    if not process.stdout.readline():
        return False
    else:
        return True

################### dialog boxes

class Information(QDialog):
    "A basic information dialog window with external links"
    def __init__(self, parent=None, title="Swac", text=''):
        QDialog.__init__(self, parent)
        self.setWindowTitle(title)
        vbox = QVBoxLayout(self)
        label = QLabel(text)
        label.setOpenExternalLinks(True)
        vbox.addWidget(label)
        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok)
        vbox.addWidget(buttonBox)
        self.connect(buttonBox, SIGNAL("accepted()"), self.accept)
        self.exec_()
    
    def accept(self):
        QDialog.accept(self)

class LangComboBox(QComboBox):
    "A combobox with the langs related to the collections installed on the system"
    def __init__(self):
        QComboBox.__init__(self)
        for langSymbol, universalLang, localLang in languagesEnabled():
            if universalLang == localLang:
                txtItem = universalLang + ' ('+langSymbol+')'
            else:
                txtItem = universalLang + ' - ' + localLang+ ' ('+langSymbol+')'
            self.addItem(txtItem, QVariant(langSymbol))

class SelectDialog(QDialog):
    "Expressions selection dialog window from Swac"
    def __init__(self):
        QDialog.__init__(self)
        self.setWindowTitle(_("Swac - Add/Suppress Swac cards"))
        height = QDesktopWidget().screenGeometry().height() * 0.8
        self.resize(248, height)
        vbox = QVBoxLayout(self)
        isSwacModel, mw.deck.currentModel = getSwacModel()
        def onChangeFunc(): pass
        modelChooser = ModelChooser(self, mw, mw.deck, onChangeFunc, cards=True)
        self.cardModelsButton = modelChooser.cards
        vbox.addWidget(self.cardModelsButton)
        self.lineEdit = QLineEdit()
        self.filterTimer = None
        self.connect(self.lineEdit, SIGNAL("textChanged(QString)"),
                        self.filterTextChanged)
        vbox.addWidget(self.lineEdit)
        self.lineEdit.setFocus()
        label = QLabel("<b>"+_("Select/Deselect expressions:")+"</b>")
        vbox.addWidget(label)
        self.countWordsSelected = 0
        self.listWidget = QListWidget()
        self.listWidget.setSelectionMode(QAbstractItemView.MultiSelection)
        self.ankiWordsLst = []
        self.swacWordsLst = []
        self.buildWordList()
        vbox.addWidget(self.listWidget)
        label = QLabel(_("There are currently <b>%(numTot)d</b> expressions "
            "in your Swac %(symbolLang)s database.<br><b>%(nomSel)s</b> "
            "expressions have been selected at the moment of the opening of this "
            "dialog window.")
                        %{'numTot':self.listWidget.count(),
                        'symbolLang':langDict[getLangTag()],
                        'nomSel': self.countWordsSelected})
        vbox.addWidget(label)
        buttonBox = QDialogButtonBox()
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Help|
                                    QDialogButtonBox.Ok)
        vbox.addWidget(buttonBox)
        self.connect(buttonBox, SIGNAL("accepted()"), self.accept)
        self.connect(buttonBox, SIGNAL("rejected()"), self.reject)
        self.connect(buttonBox, SIGNAL("helpRequested()"), self.onHelp)
    
    def buildWordList(self):
        # Swac words
        lang = getLangTag()
        self.swacWordsLst = swacGetAll("""
select SWAC_TEXT, SWAC_HOMOGRAPHIDX, SWAC_PRON_INTONATION from sounds
where SWAC_LANG='%s' ORDER BY SWAC_TEXT ASC;""" %lang)
        
        # Anki words
        modelId = mw.deck.s.column0("""
select id from models
where name='%s'""" %(config.get('swacModel')))[0]
        fieldModelId = mw.deck.s.column0("""
select id from fieldModels
where modelId=:modelId and name=:name""",
                modelId=modelId, name=config.get('swacField'))[0]
        self.ankiWordsLst = mw.deck.s.column0("""
select distinct value from fields
where fieldModelId = :fieldModelId""", fieldModelId=fieldModelId)
        
        # Selection of the Anki words
        countWordsSelected = 0
        length = len(self.swacWordsLst)
        runHook("startProgress", length, 0, _("Swac - Add/Suppress Swac cards"))
        perCentTarget = 0
        for i, (word, determinative, intonation) in enumerate(self.swacWordsLst):
            perCent = int(float(i)/length*100)
            # update the progression bar after each expression would be too long
            if perCent >= perCentTarget:
                runHook("updateProgress", _('Be patient...'), i)
                perCentTarget += 10
            if determinative:
                word += ' (%s)' %determinative
            elif intonation:
                word += ' [%s]' %intonation
            item = QListWidgetItem(word, self.listWidget)
            self.listWidget.addItem(item)
            idx = self.listWidget.indexFromItem(item)
            if word in self.ankiWordsLst:
                mode = QItemSelectionModel.Select
                countWordsSelected += 1
            else:
                mode = QItemSelectionModel.Deselect
            self.listWidget.selectionModel().select(idx, mode)
        
        runHook("updateProgress")
        runHook("finishProgress")
        self.countWordsSelected = countWordsSelected
    
    def filterTextChanged(self):
        interval = 500 # ms
        if self.filterTimer:
            self.filterTimer.setInterval(interval)
        else:
            self.filterTimer = QTimer(self)
            self.filterTimer.setSingleShot(True)
            self.filterTimer.start(interval)
            self.connect(self.filterTimer, SIGNAL("timeout()"), self.updateSearch)
    
    def updateSearch(self):
        class FoundExpression(Exception): pass
        text = unicode(self.lineEdit.text())
        try:
            if not text:
                item = self.listWidget.item(0)
                raise FoundExpression
            for i, letter in enumerate(text):
                if i:
                    ch = text[:-i]
                else:
                    ch = text
                textList = self.listWidget.findItems(ch,
                        Qt.MatchStartsWith|Qt.MatchCaseSensitive)
                if textList:
                    item = textList[0]
                    raise FoundExpression
        except FoundExpression:
            self.listWidget.scrollToItem(item, QAbstractItemView.PositionAtTop)
        finally:
            self.filterTimer = None
    
    def accept(self):
        oldAnki = set(self.ankiWordsLst)
        newAnkiList = self.listWidget.selectedItems()
        newAnki = set([unicode(i.text()) for i in newAnkiList])
        self.addWords = newAnki.difference(oldAnki)
        self.suppressWords = oldAnki.difference(newAnki)
        QDialog.accept(self)

    def reject(self):
        QDialog.reject(self)

    def onHelp(self):
        text = _(
            "Every new expression selected will create one or several new cards<br>"
            "depending on the state of the button on the top of the window.<br><br>"
            "Every expression deselected will delete the fact and all the cards <br>"
            "related (the state of button on the top doesn't matter here).<br><br>"
            "Try \"Ctrl+A\" on the expressions list if you want all select them.")
        Information(mw, _("Swac - Select expressions"), text)

def selectWords():
    "Instance SelectDialog if it's possible"
    isSwacModel, model = getSwacModel()
    if not swacSql.isConnect() or not isSwacModel or not isSwacFieldCorrect(model):
        return
    dial = SelectDialog()
    if dial.exec_():
        addWords = dial.addWords
        suppressWords = dial.suppressWords
    else:
        return
    
    if addWords:
        addFacts(addWords)
    if suppressWords:
        deleteFacts(suppressWords)
    if addWords or suppressWords:
        mw.moveToState("auto")

class Export(QDialog):
    "Export the words of a deck. Read onHelp()"
    def __init__(self):
        QDialog.__init__(self)
        self.setWindowTitle(_("Swac - Export words"))
        vbox = QVBoxLayout(self)
        label = QLabel(_("<b>Check the fields you want to export</b>"))
        vbox.addWidget(label)
        self.treeWidget = QTreeWidget()
        vbox.addWidget(self.treeWidget)
        self.treeWidget.setColumnCount(1)
        
        modelDict = {}
        for mod in mw.deck.models:
            if mod.name == config.get('swacModel'):
                continue
            fieldList = []
            for field in mod.fieldModels:
                fieldList.append(field.name)
            modelDict[mod.name] = fieldList
        
        items = []
        for model in modelDict.keys():
            itemParent = QTreeWidgetItem([model])
            items.append(itemParent)
            for field in modelDict[model]:
                item = QTreeWidgetItem(itemParent, [field])
                item.setCheckState(0, Qt.Unchecked)
                items.append(item)
        self.treeWidget.insertTopLevelItems(0, items)
        self.treeWidget.expandAll()
        
        self.comboLang = LangComboBox()
        vbox.addWidget(self.comboLang)
        
        self.comboMode = QComboBox()
        self.comboMode.addItem(_("Default export"), QVariant('defaults'))
        self.comboMode.addItem(_("Suggest export"), QVariant('special'))
        vbox.addWidget(self.comboMode)
        
        label = QLabel(_("Patterns to exclude:"))
        patternToExclude = u"\\t \\n.,;?!:()=\\+/\"'<>|[]{}«»0123456789*%"
        self.lineEdit = QLineEdit(patternToExclude)
        hbox = QHBoxLayout()
        hbox.addWidget(label)
        hbox.addWidget(self.lineEdit)
        vbox.addLayout(hbox)
        
        buttonBox = QDialogButtonBox()
        buttonBox.setStandardButtons(QDialogButtonBox.Help|QDialogButtonBox.Cancel|
                                                    QDialogButtonBox.Ok)
        vbox.addWidget(buttonBox)
        self.connect(buttonBox, SIGNAL("accepted()"), self.accept)
        self.connect(buttonBox, SIGNAL("rejected()"), self.reject)
        self.connect(buttonBox, SIGNAL("helpRequested()"), self.onHelp)
        self.exec_()
    
    def getPatternToExclude(self):
        patterns = unicode(self.lineEdit.text())
        new = ''    # new character
        prev = ''   # previous character
        
        for c in patterns:
            if prev == '\\':
                if c == 'n':
                    new += '\n'
                elif c == 't':
                    new += '\t'
                elif c == '\"':
                    new += '\"'
                else:
                    new += prev + c
            elif c == '\\':
                prev = '\\'
            else:
                new += c
            prev = c
        
        return new
    
    def simplifySentence(self, sentence=None, patternToExclude=None):
        lst = []
        if not sentence:
            return lst
        nv = ''
        sentence = stripHTML(sentence)
        for c in sentence:
            if c in patternToExclude:
                if nv:
                    lst.append(nv.lower())
                    nv = ''
            else:
                nv += c
        else:
            lst.append(nv.lower())
        return lst
    
    def export(self, words=[], fieldSeparator=';', mode='defaults'):
        path = unicode(QFileDialog.getSaveFileName(None, _(u"Save"),
                    os.path.expanduser("~"), "*.*"))
        if not path:
            return False
        try:
            f = open(path, 'w')
        except IOError, e:
            print "IOError: %s" %e
            QMessageBox.warning(self, _("Swac - Export words"),
                        _("Permission denied!"))
            return False
        for word in words:
            f.write(word.encode('utf8') + fieldSeparator + '\n')
        if mode == 'special':
            txt = _("<br><br>"
                "You can read this file, delete the words not related to<br>"
                "the language selected and send it to "
                "<a href=\"http://shtooka.net/en/contact.html\">"
                "Shtooka project</a>.<br>"
                "If you are lucky, the sounds related will be quickly add<br>"
                "to the Swac collection.")
        else:
            txt = ''
        f.close()
        Information(self, _("Swac - Export words"),
            _("%(number)s words have been copied into the file <a href="
            "\"%(path)s\">%(path)s</a>%(contact)s")
                %{'number': len(words), 'path': path, 'contact': txt})
        return True
    
    def accept(self):
        def join(txt):
            return '"' + '", "'.join(txt) + '"'
        
        modelDict = {}
        for i in range(self.treeWidget.topLevelItemCount()):
            itemParent = self.treeWidget.topLevelItem(i)
            fieldNames = []
            for j in range(itemParent.childCount()):
                item = itemParent.child(j)
                if item.checkState(0):
                    fieldNames.append(unicode(item.text(0)))
            if fieldNames:
                modelDict[unicode(itemParent.text(0))] = fieldNames
        
        if not modelDict:
            QMessageBox.warning(self, _("Swac - Export words"),
                                    _("Check at least one field to export words"))
            return
        
        lstAnkiField = []
        for modelName in modelDict.keys():
            modelId = mw.deck.s.column0("""
select id from models
where name='%s'""" %modelName)[0]
            for fieldName in modelDict[modelName]:
                fieldModelId = mw.deck.s.column0("""
select id from fieldModels
where modelId=:modelId and name=:name""", modelId=str(modelId), name=fieldName)[0]
                values = mw.deck.s.column0("""
select distinct value from fields
where fieldModelId=:fieldModelId""", fieldModelId=str(fieldModelId))
                for val in values:
                    lstAnkiField.append(val)
        
        ankiWordsLst = set()
        patternToExclude = self.getPatternToExclude()
        for sentence in lstAnkiField:
            for word in self.simplifySentence(sentence, patternToExclude):
                ankiWordsLst.add(word)
        
        if not ankiWordsLst:
            QMessageBox.warning(self, _("Swac - Export words"),
                                    _("Nothing to export"))
            QDialog.reject(self)
            return
        
        i = self.comboLang.currentIndex()
        lang = unicode(self.comboLang.itemData(i).toString())
        
        swacWordsLst = set(i[0] for i in swacGetAll("""
select SWAC_TEXT from sounds
where SWAC_LANG='%s';""" %lang))
        
        i = self.comboMode.currentIndex()
        mode = unicode(self.comboMode.itemData(i).toString())
        if mode == 'defaults':
            fieldSeparator = ';'
            words = ankiWordsLst.intersection(swacWordsLst)
        elif mode == 'special':
            fieldSeparator = ''
            words = ankiWordsLst.difference(swacWordsLst)
        if words:
            wordsOrder = [w for w in words]
            wordsOrder.sort()
            if not self.export(wordsOrder, fieldSeparator, mode):
                return
        
        QDialog.accept(self)
    
    def reject(self):
        QDialog.reject(self)
    
    def onHelp(self):
        text = _(
            "By defaults, every word found both in your fields and in<br>"
            "the Swac collection (related to the lang selected) will be export.<br>"
            "You will afterward be able to import the related file in a<br>"
            "\"%(swacField)s\" field (in this deck or another) with:<br>"
            "File menu -> Import<br><br>"
            "\"Suggest export\" makes it possible for you to send a list of <br>"
            "words to <a href=\"http://shtooka.net/en/contact.html\">Shtooka "
            "project</a>.<br>"
            "If you are lucky, the sounds related will be quickly add to <br>"
            "the Swac collection.") \
                %{'swacField': config.get('swacField')}
        Information(mw, _("Swac - Export words"), text)

class Settings(QDialog):
    "Settings dialog window. Look at onHelp. This class is very linked to configDict"
    def __init__(self):
        QDialog.__init__(self)
        vbox = QVBoxLayout(self)
        self.setWindowTitle(_("Swac Settings"))
        self.grid = QGridLayout()
        vbox.addLayout(self.grid)
        self.buildGrid()
        buttonBox = QDialogButtonBox()
        buttonBox.setStandardButtons(QDialogButtonBox.Help|QDialogButtonBox.Cancel|
                                            QDialogButtonBox.Ok)
        self.defaultsButton = buttonBox.addButton(QDialogButtonBox.RestoreDefaults)
        vbox.addWidget(buttonBox)
        self.connect(buttonBox, SIGNAL("accepted()"), self.accept)
        self.connect(buttonBox, SIGNAL("rejected()"), self.reject)
        self.connect(buttonBox, SIGNAL("helpRequested()"), self.onHelp)
        self.connect(buttonBox, SIGNAL("clicked(QAbstractButton *)"), 
                                    self.defaultParam)
        self.exec_()

    def buildGrid(self):
        for param in configDict.keys():
            setattr(self, param+'Label', QLabel(configDict[param]['label']))
            self.grid.addWidget(getattr(self, param+'Label'),
                                    configDict[param]['sort'], 0)
            setattr(self, param+'LineEdit', QLineEdit(config.get(param)))
            self.grid.addWidget(getattr(self, param+'LineEdit'),
                                    configDict[param]['sort'], 1)
    
    def updateParam(self):
        for param in configDict.keys():
            config.set(param,
                    unicode(getattr(self, param+'LineEdit').text().trimmed()))
    
    def defaultParam(self, button=None):
        if self.defaultsButton == button:
            for param in configDict.keys():
                getattr(self, param+'LineEdit').setText(configDict[param]['value'])
    
    def accept(self):
        class SettingsException(Exception): pass
        class SettingsDbException(Exception): pass
        def join(txt):
            return '"' + '", "'.join(txt) + '"'
        text = ""
        try:
            labelEmpty = []
            for param in configDict.keys():
                if not getattr(self, param+'LineEdit').text().trimmed():
                    labelEmpty.append(unicode(getattr(self, param+'Label').text()))
            if labelEmpty:
                text = _("Please, fill the field(s): %s") %join(labelEmpty)
                raise SettingsException
            fieldEditLine = []
            for param in configDict.keys():
                if 'Field' in param:
                    line = getattr(self, param+'LineEdit').text().trimmed()
                    if line in fieldEditLine:
                        text = _("The fields \"%(fieldSwac)s\", "
                        "\"%(fieldPhon)s\" and \"%(fieldHomo)s\" must be different")\
                            %{'fieldSwac':  self.swacFieldLabel.text(),
                            'fieldPhon': self.phoneticFieldLabel.text(),
                            'fieldHomo': self.homophonFieldLabel.text()}
                        raise SettingsException
                    else:
                        fieldEditLine.append(line)
            if self.readingCardModelLineEdit.text().trimmed() == \
                    self.recognitionCardModelLineEdit.text().trimmed():
                text = _("The fields \"%(readingField)s\" and \" "
                    "%(RecogField)s\" must be different")\
                        %{'readingField': self.readingCardModelLabel.text(),
                        'RecogField':  self.recognitionCardModelLabel.text()}
                raise SettingsException
            databaseDir = self.swacDatabasePathLineEdit.text().trimmed()
            if not os.path.exists(databaseDir):
                text = _("The field \"%(field)s\" is not correct. "
                "The \"%(path)s\" path doesn't exist. Would you really continue?")\
                    %{'field': self.swacDatabasePathLabel.text(),
                        'path': databaseDir}
                raise SettingsDbException
        except SettingsException:
            QMessageBox.warning(self, _("Swac settings"), text)
        except SettingsDbException:
            if QMessageBox.warning(self, _("Swac settings"), text,
                    QMessageBox.Yes | QMessageBox.No) == QMessageBox.Yes:
                self.updateParam()
                swacSql.connect(unicode(databaseDir))
                QDialog.accept(self)
        else:
            self.updateParam()
            swacSql.connect(unicode(databaseDir))
            actionSelectWordsSetShortCut()
            voiceRecording.update()
            QDialog.accept(self)
    
    def reject(self):
        QDialog.reject(self)
    
    def onHelp(self):
        text = _(
            "\"%(swacModel)s\", \"%(swacField)s\" and  <br>\"%(swacDb)s\" fields "
            "are the most important.<br>Be careful if you change them after "
            "creating a Swac model in your deck.<br>In this case you will have to "
            "manually change the \"Deck Properties\".") \
                %{'swacModel': configDict['swacModel']['label'],
                'swacField': configDict['swacField']['label'],
                'swacDb': configDict['swacDatabasePath']['label']}
        Information(mw, _("Swac - Settings"), text)

def documentation():
    "Main help"
    QDesktopServices.openUrl(QUrl(os.path.join(swacPluginResource, "swac.html")))

def informations():
    "About Swac plugin"
    QMessageBox.information(mw, _("About Swac (plugin)"),
        _("""<h2>Swac Plugin<br>%(version)s version ; %(date)s</h2>
<p>Swac plugin is designed to memorize sounds words of the Swac collections.
It has been tested on Anki 0.9.9.6</p>

<p>Swac plugin is based on the Swac tools swac-get and swac-play.</p>

<p>Both plugin and Swac tools are free and licensed under the GPL3.</p>

<p>For more information about Swac tools, please, visit the Shtooka Project 
homepage at: <a href="http://shtooka.net/">http://shtooka.net/</a>.<br>
For more information about Swac plugin, please,
<a href="%(helpPath)s">look at the documentation</a></p>

<p><a href=
"http://ichi2.net/anki/wiki/Plugins?action=\
AttachFile&do=view&target=anki_swac.tar.bz2">
Visit the Anki website to get the last version of the plugin</a></p>

<p>Swac plugin author: Alteo Gange <a href="maito:romanocaldoni@free.fr">
romanocaldoni@free.fr</a></p>
""") %{ 'version': __version__, 'date': __date__,
        'helpPath': os.path.join(swacPluginResource, _("swac.html"))})

class SelectLang(QDialog):
    """Selection of the language of the Swac model (related to the collections
    installed on the system)"""
    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        self.setWindowTitle("Swac")
        vbox = QVBoxLayout(self)
        buttonBox = QDialogButtonBox()
        buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)
        label = QLabel(_(u"Choose a language for the audio (Swac) model:"))
        self.combo = LangComboBox()
        vbox.addWidget(label)
        vbox.addWidget(self.combo)
        vbox.addStretch()
        vbox.addWidget(buttonBox)
        self.connect(buttonBox, SIGNAL("accepted()"), self.accept)
        self.connect(buttonBox, SIGNAL("rejected()"), self.reject)

    def accept(self):
        i = self.combo.currentIndex()
        self.lang = unicode(self.combo.itemData(i).toString())
        QDialog.accept(self)

    def reject(self):
        QDialog.reject(self)

################### Swac database

class SwacDataBaseConnect():
    "This class implements a basic handling for the Swac database + sqlite3"
    def __init__(self):
        self.connexion = None
        self.cursor = None
        self.connect()
    
    def connect(self, path=None):
        if not path:
            path = config.get('swacDatabasePath')
        if self.connexion:
            self.connexion.close()
        if os.path.exists(path):
            self.connexion = sqlite3.connect(config.get('swacDatabasePath'))
            self.cursor = self.connexion.cursor()
            config.set('swacDatabasePath', path)
        else:
            self.connexion = None
            self.cursor = None
    
    def isConnect(self):
        return self.connexion
    
    def execute(self, sql):
        if self.isConnect():
            return self.cursor.execute(sql)
        else:
            return False

swacSql = SwacDataBaseConnect()
if not swacSql.isConnect():
    swacDataBaseError()

def swacGet(sql):
    "Get Swac the audio expressions list related to a language"
    query = swacSql.execute(sql).fetchall()
    return query
    #sqlite3.OperationalError

def swacGetOne(sql):
    lst = swacGet(sql)
    if lst:
        return [i[0] for i in lst]

def swacGetAll(sql):
    return swacGet(sql)

################### audio functions and class

class VoiceRecording(object):
    """Record your voice and compare it to the native speaker.
    It is inpired buy "Voice Recording to improve pronunciation" plugin
    Thanks to Pal Dorogi"""
    def __init__(self):
        self.__voiceAudio = None
        self.__playSwac = False

        self.recordVoice = QAction(mw)
        self.replayVoice = QAction(mw)
        self.recordVoiceShortcut = QShortcut(mw)
        self.replayVoiceShortcut = QShortcut(mw)
        mw.connect(self.recordVoice, SIGNAL('triggered()'), self.onRecordVoice)
        mw.connect(self.replayVoice, SIGNAL('triggered()'), self.onReplayVoice)
        mw.connect(self.recordVoiceShortcut, SIGNAL('activated()'), self.onRecordVoice)
        mw.connect(self.replayVoiceShortcut, SIGNAL('activated()'), self.onReplayVoice)
        self.recordVoice.setEnabled(True)
        self.replayVoice.setEnabled(False)

        self.separator = mw.mainWin.toolBar.addSeparator()

        icon = QIcon()
        icon.addFile(os.path.join(swacPluginResource,"media-record-blue.png"))
        self.recordVoice.setIcon(icon)
        icon.addFile(os.path.join(swacPluginResource,"media-playback-start-blue.png"))
        self.replayVoice.setIcon(icon)
        
        mw.mainWin.toolBar.addAction(self.recordVoice)
        mw.mainWin.toolBar.addAction(self.replayVoice)
        
        self.update()

        self.recordVoice.setObjectName("actionRecordVoice")
        self.replayVoice.setObjectName("actionReplayVoice")

    def update(self):
        recordShortcut = config.get('recordShortcut')
        replayShortcut = config.get('replayShortcut')
        recordMessage = _('Record your voice (%s)') %recordShortcut
        replayMessage = _('Repeat your voice (%s)') %replayShortcut
        
        self.recordVoice.setStatusTip(recordMessage)
        self.replayVoice.setStatusTip(replayMessage)
        self.recordVoice.setText(recordMessage)
        self.replayVoice.setText(replayMessage)
        
        self.recordVoiceShortcut.setKey(recordShortcut)
        self.replayVoiceShortcut.setKey(replayShortcut)
        
        if config.get('viewPlayIcons') == "0":
            boolean = True
        else:
            boolean = False
        self.separator.setVisible(boolean)
        self.recordVoice.setVisible(boolean)
        self.replayVoice.setVisible(boolean)

    def noAddToolBar(self):
        mw.addAction(self.recordVoice)
        mw.addAction(self.replayVoice)

    def onRecordVoice(self):
        self.__voiceAudio = getAudio(mw.mainWin.toolBar)
        self.replayVoice.setEnabled(True)
        # see "play('onAnswer')" line
        self.setEnabledSwac(True)

    def onReplayVoice(self):
        clearAudioQueue()
        playOrigin(self.__voiceAudio)

    def setEnabledSwac(self, enable=False):
        self.__playSwac = enable

    def playSwac(self):
        return self.__playSwac

def swacPlay(expression, homograph=None, intonation=None, lang=None, wait=True):
    "Play a word (based on gstreamer0.10)"
    command = ["swac-play"]
    if lang:
        command.append("--lang=" + lang)
    if homograph:
        command.append("--homographidx=" + homograph)
    if intonation:
        command.append("--intonation=" + intonation)
    command.append("\""+expression+"\"")
    process = subprocess.Popen(" ".join(command), stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT, shell=True)
    if wait:
        process.wait() # will return true or false only after the end of the command
    try:
        log = process.stdout.readline()
    except IOError:
        return False, None
    if "Can not find matching audio record" in log:
        return False, 'notFile'
    elif "not found" in log:
        if not isBinary("swac-play"):
            return False, 'notBinary'
    else:
        return True, None

def play(qr):
    "Play an expression if it's possible or return False"
    currentModel = mw.currentCard.fact.model.name
    if currentModel != config.get('swacModel'):
        return False
    if not swacSql.isConnect():
        return False
    symbolLang = getLangTag()
    if not symbolLang:
        return False
    try:
        value = mw.currentCard.fact[config.get('swacField')]
    except KeyError:
        swacFieldError()
        return False
    if not value:
        return False
    qformat = mw.currentCard.cardModel.qformat
    aformat = mw.currentCard.cardModel.aformat
    swacField = '%('+config.get('swacField')+')s'
    toPlay = False
    if qr == 'onQuestion':
        if swacField not in qformat and swacField in aformat:
            toPlay = True
    elif qr == 'onAnswer':
        if swacField in qformat and swacField not in aformat:
            toPlay = True
    elif qr == 'onAny':
        if swacField in qformat or swacField in aformat:
            toPlay = True
    if toPlay:
        expression, determinative, stressed = splitSwacPattern(value)
        if qr == 'onAnswer':
            waiting = True
        else:
            waiting = False
        isPlay, msg = swacPlay(expression, homograph=determinative,
                                intonation=stressed, lang=symbolLang, wait=waiting)
        if isPlay:
            return True
        else:
            if msg == 'notBinary':
                swacBinaryError("swac-play")
            elif msg == 'notFile':
                try:
                    phonVal = mw.currentCard.fact[config.get('phoneticField')]
                except KeyError:
                    phonVal = ''
                if not phonVal:
                    findAudioError(value)
            return False
    else:
        if swacField not in qformat and swacField in aformat:
            ENABLE_PLAY_BUTTON = True
        else:
            ENABLE_PLAY_BUTTON = False
    return False

################### deck, fact and model functions

def newSwacModel():
    """The Swac model will appear in: menu Settings -> Deck proporties
    -> Models and properties tab -> add button -> Add model window"""
    m = stdmodels.Model(_(config.get('swacModel')))
    m.addFieldModel(stdmodels.FieldModel(config.get('swacField'), True, True))
    m.addFieldModel(stdmodels.FieldModel(config.get('phoneticField'), False, False))
    m.addFieldModel(stdmodels.FieldModel(config.get('homophonField'), False, False))
    swacField = '%(' + config.get('swacField') + ')s'
    phoneticField = '<span style="font-size:0.65em; font-weight:bold;">%(' + \
                config.get('phoneticField') + ')s</span>'
    homophonField = '%(' + config.get('homophonField') + ')s'
    tagField = '<span style="color:grey;font-size:0.75em;">' + """%(Tags)s""" + \
                    '</span>'
    iconPath = u"file://" + os.path.join(swacPluginResource, "text-speak.png")
    m.addCardModel(stdmodels.CardModel(config.get('readingCardModel'),
                    _("Pronounce: %s") %swacField,
                    '<img src="%s">\n' %iconPath + phoneticField + '\n' + tagField))
    cardModel = stdmodels.CardModel(config.get('recognitionCardModel'),
                    '<img src="%s">' %iconPath,
                    swacField + '\n' + homophonField + '\n' + phoneticField  + '\n' + tagField)
    cardModel.typeAnswer = config.get('swacField')
    m.addCardModel(cardModel)
    m.tags = config.get('swacModel')
    return m

stdmodels.swacModel = newSwacModel
stdmodels.models[config.get('swacModel')] = stdmodels.swacModel

def getLangTag():
    "Return the ISO-639-3 symbol language of your Swac model or False"
    modelTags = mw.deck.s.column0("""
select tags from models
where name='%s'""" %(config.get('swacModel')))[0]
    modelTagsList = parseTags(modelTags)
    langSymbolList = swacGetOne("select iso_639_3 from languages;")
    if not langSymbolList:
        return False
    langTags = []
    for tag in langSymbolList:
        if findTag(tag, modelTagsList):
            langTags.append(tag)
    lenLangTags = len(langTags)
    if lenLangTags == 1:
        return langTags[0]
    elif lenLangTags == 0:
        langAnkiError()
        return False
    elif lenLangTags > 1:
        langsAnkiError(langTags)
        return False

def getSwacModel():
    "Return a tuple boolean, Swac model instance"
    models = []
    for model in mw.deck.models:
        if model.name == config.get('swacModel'):
            models.append(model)
    lenModel = len(models)
    if lenModel == 1:
        return True, models[0]
    elif lenModel > 1:
        severalSwacModelError()
        return False, None
    elif lenModel == 0: 
        return False, None

def isSwacFieldCorrect(model):
    "Return True if one Swac field exists"
    if not model:
        return False
    fields = []
    for field in model.fieldModels:
        if field.name == config.get('swacField'):
            fields.append(field)
    lenField = len(fields)
    if lenField == 1:
        return True
    elif lenField > 1:
        severalSwacFieldError()
        return False
    elif lenField == 0:
        swacFieldError()
        return False

def addFacts(words):
    """Add fact -> add cards related to the state of the question/answer button
    Look at SelectDialog.onHelp"""
    for w in words:
        fact = mw.deck.newFact()
        for field in fact.fields:
            if field.name == config.get('swacField'):
                field.value = w
        mw.deck.addFact(fact)

def deleteFacts(words):
    """Deleted facts (and all the cards related). Look at SelectDialog.onHelp().
    The facts whose at least two fields are not empty won't be deleted. It's
    especially useful if you have created a fact with a full phonetic field."""
    def join(words, id=False):
        if id:
            lst = [str(i.id) for i in words]
        else:
            lst = words
        return "\'" + "\', \'".join(lst) + "\'"
    
    swacFieldModel, otherFieldsModel = None, []
    isSwacModel, swacModel = getSwacModel()
    for fieldModel in swacModel.fieldModels:
        if fieldModel.name == config.get('swacField'):
            swacFieldModel = fieldModel
        else:
            otherFieldsModel.append(fieldModel)
    cardModels = swacModel.cardModels
    cardsIdList = []
    cardsValueList = []
    
    for word in words:
        otherFieldsModelAreEmpty = True
        factId = mw.deck.s.column0("""
select factId from fields
where value = '%s' and fieldModelId = '%s'""" %(word, swacFieldModel.id))[0]
        for fieldModel in otherFieldsModel:
            value = mw.deck.s.column0("""
select value from fields
where factId = '%s' and fieldModelId = '%s'""" %(factId, fieldModel.id))[0]
            if value:
                otherFieldsModelAreEmpty = False
                break
        if otherFieldsModelAreEmpty:
            cardsId = mw.deck.s.column0("""
select id from cards
where factId = '%s' and cardModelId in (%s)""" %(factId, join(cardModels,id=1)))
            for c in cardsId:
                cardsIdList.append(c)
            cardsValueList.append(word)
    
    if not cardsIdList:
        return
    if QMessageBox.warning(mw, _("Swac words"),
                        _("The followings words will be deleted:\n%s\n\n"
                        "Do you want to apply this change?") %join(cardsValueList),
                        QMessageBox.Yes | QMessageBox.No) == QMessageBox.Yes:
        mw.deck.deleteCards(cardsIdList)

################### functions and signals interceptions of Anki

def newShowEaseButtons():
    oldShowEaseButtons()
    QTimer.singleShot(250, newShowEaseButtons2)

def newShowEaseButtons2():
    global ENABLE_PLAY_BUTTON
    if play('onAnswer'):
        ENABLE_PLAY_BUTTON = True
        mw.mainWin.actionRepeatAudio.setEnabled(True)
        if voiceRecording.playSwac():
            voiceRecording.onReplayVoice()

def newEnableCardMenuItems():
    global ENABLE_PLAY_BUTTON
    oldEnableCardMenuItems()
    if ENABLE_PLAY_BUTTON:
        mw.mainWin.actionRepeatAudio.setEnabled(True)
        ENABLE_PLAY_BUTTON = False

def newDisableCardMenuItems():
    global ENABLE_PLAY_BUTTON
    oldDisableCardMenuItems()
    mw.mainWin.actionRepeatAudio.setEnabled(False)
    ENABLE_PLAY_BUTTON = False

def newOnRepeatAudio():
    if not play('onAny'):
        mw.onRepeatAudio()

def onShowQuestion():
    global ENABLE_PLAY_BUTTON
    voiceRecording.setEnabledSwac(False)
    ENABLE_PLAY_BUTTON = play('onQuestion')
    if ENABLE_PLAY_BUTTON:
        mw.mainWin.actionRepeatAudio.setEnabled(True)

def newAcceptAddModel(parent=None):
    nameModelAdd = parent.models[parent.dialog.models.currentRow()][1].name
    if nameModelAdd == config.get('swacModel'):
        dialog = SelectLang(mw.deckProperties)
        if dialog.exec_():
            lang = dialog.lang
            swacModel = parent.models[parent.dialog.models.currentRow()][1]
            swacModel.tags = addTags(lang, swacModel.tags)
            oldAcceptAddModel(parent)
        else:
            AddModel.reject(parent)
    else:
        oldAcceptAddModel(parent)

def newCloseEvent(parent=None):
    try:
        f = open(configFilePath,'w')
        config.write(f)
    except IOError, e:
        print "IOError: %s" %e
    oldCloseEvent(parent)

oldShowEaseButtons = mw.showEaseButtons
mw.showEaseButtons = newShowEaseButtons

oldEnableCardMenuItems = mw.enableCardMenuItems
mw.enableCardMenuItems = newEnableCardMenuItems

oldDisableCardMenuItems = mw.disableCardMenuItems
mw.disableCardMenuItems = newDisableCardMenuItems

oldAcceptAddModel = AddModel.accept
AddModel.accept = newAcceptAddModel

oldCloseEvent = mw.closeEvent
mw.closeEvent = newCloseEvent

mw.disconnect(mw.mainWin.actionRepeatAudio, SIGNAL("triggered()"), mw.onRepeatAudio)
mw.connect(mw.mainWin.actionRepeatAudio, SIGNAL("triggered()"), newOnRepeatAudio)

mw.addHook("showQuestion", onShowQuestion)

################### new actions (menu)

if hasattr(mw.mainWin, 'menuPlugins'):
    parentMenu = getattr(mw.mainWin, 'menuPlugins')
else:
    parentMenu = getattr(mw.mainWin, 'menu_Settings')
parentMenu.addSeparator()
swacMenu = parentMenu.addMenu(_("&Swac Settings"))

actionSelectWords = QAction(_("&Add/Suppress Swac cards"), mw)
def actionSelectWordsSetShortCut():
    actionSelectWords.setShortcut(config.get('swacWordsSelectShortcut'))
actionSelectWordsSetShortCut()
swacMenu.addAction(actionSelectWords)
mw.connect(actionSelectWords, SIGNAL("triggered()"), selectWords)
 
actionExport = QAction(_("&Export"), mw)
swacMenu.addAction(actionExport)
mw.connect(actionExport, SIGNAL("triggered()"), Export)

actionSwacSettings = QAction(_("&Settings"), mw)
swacMenu.addAction(actionSwacSettings)
mw.connect(actionSwacSettings, SIGNAL("triggered()"), Settings)

swacMenu.addSeparator()

actionSwacDoc = QAction(_("&Help"), mw)
swacMenu.addAction(actionSwacDoc)
mw.connect(actionSwacDoc, SIGNAL("triggered()"), documentation)

actionSwacInformations = QAction(_("A&bout..."), mw)
swacMenu.addAction(actionSwacInformations)
mw.connect(actionSwacInformations, SIGNAL("triggered()"), informations)

def enableActions():
    "Enable/disable actions in Swac Settings menu"
    class NotDatabase(Exception): pass
    class NotSwacModel(Exception): pass
    class NotSwacField(Exception): pass
    if mw.state == "noDeck":
        swacMenu.setEnabled(False)
        return
    else:
        swacMenu.setEnabled(True)
    try:
        if not swacSql.isConnect():
            raise NotDatabase
        isSwacModel, model = getSwacModel()
        if not isSwacModel:
            raise NotSwacModel
        isSwacField = isSwacFieldCorrect(model)
        if not isSwacField:
            raise NotSwacField
    except NotDatabase:
        enableSelectWords = False
        enableExport = False
    except (NotSwacModel, NotSwacField):
        enableSelectWords = False
        enableExport = True
    else:
        enableSelectWords = True
        enableExport = True
    finally:
        if not swacSql.isConnect():
            swacDataBaseError()
        actionSelectWords.setEnabled(enableSelectWords)
        actionExport.setEnabled(enableExport)

mw.connect(parentMenu, SIGNAL("aboutToShow()"), enableActions)

voiceRecording = VoiceRecording()

# update notification
if appVersion >= "0.9.9.7":
	mw.registerPlugin(_("Swac Audio Collections"), 214)
